#[derive(Eq, PartialEq, Copy, Clone, Debug)]
pub enum Color {
    Red,
    Green,
    Yellow,
    Blue,
}

#[derive(Eq, PartialEq, Copy, Clone, Debug)]
pub enum Resource {
    Brick,
    Wood,
    Sheep,
    Wheat,
    Ore,
}

#[derive(Eq, PartialEq, Copy, Clone, Debug)]
pub enum DevCard {
    Knight,
    VictoryPoint,
    RoadBuilding,
    YearOfPlenty,
    Monopoly,
}

#[derive(Eq, PartialEq, Copy, Clone, Debug)]
pub enum HexType {
    Desert,
    Resource,
    Port,
    Ocean,
}

/**
* Skewed 2d coordinate system
*
*    \
*     \
* -----\-----> y
*       \ 60 deg
*        \ x
*/
#[derive(Eq, Ord, PartialEq, PartialOrd, Copy, Clone, Debug)]
pub struct Position(pub i32, pub i32);

#[derive(Eq, PartialEq, Copy, Clone, Debug)]
pub struct Settlement(pub Color);
#[derive(Eq, PartialEq, Copy, Clone, Debug)]
pub struct City(pub Color);
#[derive(Eq, PartialEq, Copy, Clone, Debug)]
pub struct Road(pub Color);

pub struct Player {
    pub name: String,
    pub color: Color,
    pub resources: Vec<Resource>,
    pub dev_cards: Vec<DevCard>,
    pub settlements: u32,
    pub cities: u32,
    pub roads: u32,
    pub knights: u32,
    pub points: u32,
    pub ports: Vec<Resource>,
}
#[derive(Eq, PartialEq, Copy, Clone, Debug)]
pub struct Hex {
    pub hextype: HexType,
    pub resource: Option<Resource>,
    pub position: Position,
    pub number: Option<u32>,
    pub orientation: Option<Position>, // orientation of the port
}

#[derive(Eq, PartialEq, Copy, Clone, Debug)]
pub struct Edge {
    pub position: Position, // half scale coordinates
    pub road: Option<Road>,
}

#[derive(Eq, PartialEq, Copy, Clone, Debug)]
pub struct Vertex {
    pub position: Position, // rotated 30 deg, half scale coordinates
    pub settlement: Option<Settlement>,
    pub city: Option<City>,
}

#[derive(Debug)]
pub struct Board {
    pub hexes: [[Option<Hex>; 7]; 7], // center is 3,3
    pub edges: [[Option<Edge>; 15]; 15],
    pub vertices: [[Option<Vertex>; 15]; 15],
    pub robber: Position,
}

pub enum GamePhase {
    Setup,
    Main,
}

pub struct Game {
    pub players: Vec<Player>,
    pub board: Board,
    pub phase: GamePhase,
    pub turn: Player,
    pub dev_cards: Vec<DevCard>,
    pub longest_road: Option<Player>,
    pub larget_army: Option<Player>,
}

impl Position {
    pub fn new(x: i32, y: i32) -> Position {
        Position(x, y)
    }

    pub fn distance(&self, other: &Position) -> u32 {
        let a = self.0 - other.0;
        let b = self.1 - other.1;
        if a ^ b >= 0 {
            (a.abs() + b.abs()) as u32
        } else {
            (a.abs().max(b.abs())) as u32
        }
    }

    pub fn rotate(&mut self) -> &mut Position {
        let x = self.0;
        let y = self.1;
        self.0 = -y;
        self.1 = x + y;
        self
    }

    pub fn add(&mut self, other: &Position) -> &mut Position {
        self.0 += other.0;
        self.1 += other.1;
        self
    }

    pub fn neg(&mut self) -> &mut Position {
        self.0 = -self.0;
        self.1 = -self.1;
        self
    }

    pub fn scale(&mut self, n: i32) -> &mut Position {
        self.0 *= n;
        self.1 *= n;
        self
    }

    pub fn edge(&self, direction: &Position) -> Position {
        let mut p = *self;
        p.scale(2).add(&Position(1, 1)).add(direction);
        p
    }

    pub fn vertex(&self, direction: &Position) -> Position {
        let Position(x, y) = *self;
        Position(2 * x + y - 2 + direction.0, y - x + 7 + direction.1)
    }
}
